@@ -225,6 +225,7 @@ var BuiltinGlobalVariableTypes = map[string]ExprType{
 		"run_id":              StringType{},
 		"run_number":          StringType{},
 		"run_attempt":         StringType{},
+		"secret_source":       StringType{},
 		"server_url":          StringType{},
 		"sha":                 StringType{},
 		"token":               StringType{},
@@ -255,12 +256,13 @@ var BuiltinGlobalVariableTypes = map[string]ExprType{
 	"steps": NewEmptyStrictObjectType(), // This value will be updated contextually
 	// https://docs.github.com/en/actions/learn-github-actions/contexts#runner-context
 	"runner": NewStrictObjectType(map[string]ExprType{
-		"name":       StringType{},
-		"os":         StringType{},
-		"arch":       StringType{},
-		"temp":       StringType{},
-		"tool_cache": StringType{},
-		"debug":      StringType{},
+		"name":        StringType{},
+		"os":          StringType{},
+		"arch":        StringType{},
+		"temp":        StringType{},
+		"tool_cache":  StringType{},
+		"debug":       StringType{},
+		"environment": StringType{}, // https://github.com/github/docs/issues/32443
 	}),
 	// https://docs.github.com/en/actions/learn-github-actions/contexts#secrets-context
 	"secrets": NewMapObjectType(StringType{}),
@@ -299,16 +301,18 @@ type ExprSemanticsChecker struct {
 	untrusted             *UntrustedInputChecker
 	availableContexts     []string
 	availableSpecialFuncs []string
+	configVars            []string
 }
 
 // NewExprSemanticsChecker creates new ExprSemanticsChecker instance. When checkUntrustedInput is
 // set to true, the checker will make use of possibly untrusted inputs error.
-func NewExprSemanticsChecker(checkUntrustedInput bool) *ExprSemanticsChecker {
+func NewExprSemanticsChecker(checkUntrustedInput bool, configVars []string) *ExprSemanticsChecker {
 	c := &ExprSemanticsChecker{
 		funcs:           BuiltinFuncSignatures,
 		vars:            BuiltinGlobalVariableTypes,
 		varsCopied:      false,
 		githubVarCopied: false,
+		configVars:      configVars,
 	}
 	if checkUntrustedInput {
 		c.untrusted = NewUntrustedInputChecker(BuiltinUntrustedInputs)
@@ -396,7 +400,14 @@ func (sema *ExprSemanticsChecker) UpdateSecrets(ty *ObjectType) {
 // UpdateInputs updates 'inputs' context object to given object type.
 func (sema *ExprSemanticsChecker) UpdateInputs(ty *ObjectType) {
 	sema.ensureVarsCopied()
-	sema.vars["inputs"] = ty
+	o := sema.vars["inputs"].(*ObjectType)
+	if len(o.Props) == 0 && o.IsStrict() {
+		sema.vars["inputs"] = ty
+		return
+	}
+	// When both `workflow_call` and `workflow_dispatch` are the triggers of the workflow, `inputs` context can be used
+	// by both events. To cover both cases, merge `inputs` contexts into one object type. (#263)
+	sema.vars["inputs"] = o.Merge(ty)
 }
 
 // UpdateDispatchInputs updates 'github.event.inputs' and 'inputs' objects to given object type.
@@ -500,6 +511,12 @@ func (sema *ExprSemanticsChecker) checkSpecialFunctionAvailability(n *FuncCallNo
 	)
 }
 
+func (sema *ExprSemanticsChecker) visitUntrustedCheckerOnEnterNode(n ExprNode) {
+	if sema.untrusted != nil {
+		sema.untrusted.OnVisitNodeEnter(n)
+	}
+}
+
 func (sema *ExprSemanticsChecker) visitUntrustedCheckerOnLeaveNode(n ExprNode) {
 	if sema.untrusted != nil {
 		sema.untrusted.OnVisitNodeLeave(n)
@@ -530,6 +547,9 @@ func (sema *ExprSemanticsChecker) checkObjectDeref(n *ObjectDerefNode) ExprType
 			return t
 		}
 		if ty.Mapped != nil {
+			if v, ok := n.Receiver.(*VariableNode); ok && v.Name == "vars" {
+				sema.checkConfigVariables(n)
+			}
 			return ty.Mapped
 		}
 		if ty.IsStrict() {
@@ -571,6 +591,56 @@ func (sema *ExprSemanticsChecker) checkObjectDeref(n *ObjectDerefNode) ExprType
 	}
 }
 
+func (sema *ExprSemanticsChecker) checkConfigVariables(n *ObjectDerefNode) {
+	// https://docs.github.com/en/actions/learn-github-actions/variables#naming-conventions-for-configuration-variables
+	if strings.HasPrefix(n.Property, "github_") {
+		sema.errorf(
+			n,
+			"configuration variable name %q must not start with the GITHUB_ prefix (case insensitive). note: see the convention at https://docs.github.com/en/actions/learn-github-actions/variables#naming-conventions-for-configuration-variables",
+			n.Property,
+		)
+		return
+	}
+	for _, r := range n.Property {
+		// Note: `n.Property` was already converted to lower case by parser
+		// Note: First character cannot be number, but it was already checked by parser
+		if '0' <= r && r <= '9' || 'a' <= r && r <= 'z' || r == '_' {
+			continue
+		}
+		sema.errorf(
+			n,
+			"configuration variable name %q can only contain alphabets, decimal numbers, and '_'. note: see the convention at https://docs.github.com/en/actions/learn-github-actions/variables#naming-conventions-for-configuration-variables",
+			n.Property,
+		)
+		return
+	}
+
+	if sema.configVars == nil {
+		return
+	}
+	if len(sema.configVars) == 0 {
+		sema.errorf(
+			n,
+			"no configuration variable is allowed since the variables list is empty in actionlint.yaml. you may forget adding the variable %q to the list",
+			n.Property,
+		)
+		return
+	}
+
+	for _, v := range sema.configVars {
+		if strings.EqualFold(v, n.Property) {
+			return
+		}
+	}
+
+	sema.errorf(
+		n,
+		"undefined configuration variable %q. defined configuration variables in actionlint.yaml are %s",
+		n.Property,
+		sortedQuotes(sema.configVars),
+	)
+}
+
 func (sema *ExprSemanticsChecker) checkArrayDeref(n *ArrayDerefNode) ExprType {
 	switch ty := sema.check(n.Receiver).(type) {
 	case AnyType:
@@ -716,11 +786,11 @@ func checkFuncSignature(n *FuncCallNode, sig *FuncSignature, args []ExprType) *E
 	return nil
 }
 
-func (sema *ExprSemanticsChecker) checkBuiltinFunctionCall(n *FuncCallNode, sig *FuncSignature) {
+func (sema *ExprSemanticsChecker) checkBuiltinFunctionCall(n *FuncCallNode, _ *FuncSignature) {
 	sema.checkSpecialFunctionAvailability(n)
 
 	// Special checks for specific built-in functions
-	switch n.Callee {
+	switch strings.ToLower(n.Callee) {
 	case "format":
 		lit, ok := n.Args[0].(*StringNode)
 		if !ok {
@@ -794,22 +864,121 @@ func (sema *ExprSemanticsChecker) checkNotOp(n *NotOpNode) ExprType {
 	return BoolType{}
 }
 
+func validateCompareOpOperands(op CompareOpNodeKind, l, r ExprType) bool {
+	// Comparison behavior: https://docs.github.com/en/actions/learn-github-actions/expressions#operators
+	switch op {
+	case CompareOpNodeKindEq, CompareOpNodeKindNotEq:
+		switch l := l.(type) {
+		case AnyType, NullType:
+			return true
+		case NumberType, BoolType, StringType:
+			switch r.(type) {
+			case *ObjectType, *ArrayType:
+				// These are coerced to NaN hence the comparison result is always false
+				return false
+			default:
+				return true
+			}
+		case *ObjectType:
+			switch r.(type) {
+			case *ObjectType, NullType, AnyType:
+				return true
+			default:
+				return false
+			}
+		case *ArrayType:
+			switch r := r.(type) {
+			case *ArrayType:
+				return validateCompareOpOperands(op, l.Elem, r.Elem)
+			case NullType, AnyType:
+				return true
+			default:
+				return false
+			}
+		default:
+			panic("unreachable")
+		}
+	case CompareOpNodeKindLess, CompareOpNodeKindLessEq, CompareOpNodeKindGreater, CompareOpNodeKindGreaterEq:
+		// null, bool, array, and object cannot be compared with these operators
+		switch l.(type) {
+		case AnyType, NumberType, StringType:
+			switch r.(type) {
+			case NullType, BoolType, *ObjectType, *ArrayType:
+				return false
+			default:
+				return true
+			}
+		case NullType, BoolType, *ObjectType, *ArrayType:
+			return false
+		default:
+			panic("unreachable")
+		}
+	default:
+		return true
+	}
+}
+
 func (sema *ExprSemanticsChecker) checkCompareOp(n *CompareOpNode) ExprType {
-	sema.check(n.Left)
-	sema.check(n.Right)
-	// Note: Comparing values is very loose. Any value can be compared with any value without an
-	// error.
-	// https://docs.github.com/en/actions/learn-github-actions/expressions#operators
+	l := sema.check(n.Left)
+	r := sema.check(n.Right)
+
+	if !validateCompareOpOperands(n.Kind, l, r) {
+		sema.errorf(n, "%q value cannot be compared to %q value with %q operator", l.String(), r.String(), n.Kind.String())
+	}
+
 	return BoolType{}
 }
 
+// checkWithNarrowing checks type of given expression with type narrowing. Type narrowing narrows
+// down the type of the expression by assuming its value. For example, `l && r` is typed as
+// `typeof(l) | typeof(r)` usually. However when the expression is assumed to be true, its type can
+// be narrowed down to `typeof(r)`.
+// This analysis is useful to make type checking more accurate. For example, `some_var && 60 || 20`
+// can be typed as `number` instead of `typeof(some_var) | number`. (#384)
+func (sema *ExprSemanticsChecker) checkWithNarrowing(n ExprNode, isTruthy bool) ExprType {
+	switch n := n.(type) {
+	case *LogicalOpNode:
+		switch n.Kind {
+		case LogicalOpNodeKindAnd:
+			// When `l && r` is true, narrow its type to `typeof(r)`
+			if isTruthy {
+				sema.check(n.Left)
+				return sema.check(n.Right)
+			}
+		case LogicalOpNodeKindOr:
+			// When `l || r` is false, narrow its type to `typeof(r)`
+			if !isTruthy {
+				sema.check(n.Left)
+				return sema.check(n.Right)
+			}
+		}
+		return sema.checkLogicalOp(n)
+	case *NotOpNode:
+		return sema.checkWithNarrowing(n.Operand, !isTruthy)
+	default:
+		return sema.check(n)
+	}
+}
+
 func (sema *ExprSemanticsChecker) checkLogicalOp(n *LogicalOpNode) ExprType {
-	lty := sema.check(n.Left)
-	rty := sema.check(n.Right)
-	return lty.Merge(rty)
+	switch n.Kind {
+	case LogicalOpNodeKindAnd:
+		// When `l` is false in `l && r`, its type is `typeof(l)`. Otherwise `typeof(r)`.
+		// Narrow the type of LHS expression by assuming its value is falsy.
+		return sema.checkWithNarrowing(n.Left, false).Merge(sema.check(n.Right))
+	case LogicalOpNodeKindOr:
+		// When `l` is true in `l || r`, its type is `typeof(l)`. Otherwise `typeof(r).
+		// Narrow the type of LHS expression by assuming its value is truthy.
+		return sema.checkWithNarrowing(n.Left, true).Merge(sema.check(n.Right))
+	default:
+		sema.check(n.Left)
+		sema.check(n.Right)
+		return AnyType{}
+	}
 }
 
 func (sema *ExprSemanticsChecker) check(expr ExprNode) ExprType {
+	sema.visitUntrustedCheckerOnEnterNode(expr)
 	defer sema.visitUntrustedCheckerOnLeaveNode(expr) // Call this method in bottom-up order
 
 	switch e := expr.(type) {
