@@ -1,6 +1,7 @@
 package actionlint
 
 import (
+	"fmt"
 	"strings"
 	"testing"
 	"unicode"
@@ -22,6 +23,7 @@ func TestExprSemanticsCheckOK(t *testing.T) {
 		jobs          *ObjectType
 		availContexts []string
 		availSPFuncs  []string
+		configVars    []string
 	}{
 		{
 			what:     "null",
@@ -493,7 +495,7 @@ func TestExprSemanticsCheckOK(t *testing.T) {
 			expected: StringType{},
 		},
 		{
-			what:     "format() function arguments varlidation",
+			what:     "format() function arguments validation",
 			input:    "format('{0}{0}{0} {1}{2}{1} {1}{2}{1}{2} {0} {1}{1}{1} {2}{2}{2} {0}{0}{0}{0} {0}', 1, 'foo', true)",
 			expected: StringType{},
 		},
@@ -645,6 +647,78 @@ func TestExprSemanticsCheckOK(t *testing.T) {
 			expected:     StringType{},
 			availSPFuncs: []string{"hashfiles"},
 		},
+		{
+			what:     "configuration variable",
+			input:    "vars.SOME_VARIABLE",
+			expected: StringType{},
+		},
+		{
+			what:       "known configuration variable",
+			input:      "vars.SOME_VARIABLE",
+			expected:   StringType{},
+			configVars: []string{"SOME_VARIABLE"},
+		},
+		{
+			what:       "configuration variable name is case insensitive",
+			input:      "vars.SOME_VARIABLE",
+			expected:   StringType{},
+			configVars: []string{"some_variable"},
+		},
+		{
+			what:     "narrow type of && operator by assumed value (#384)",
+			input:    "('foo' && 10) || 20",
+			expected: NumberType{},
+		},
+		{
+			what:     "narrow type of || operator by assumed value (#384)",
+			input:    "('foo' || 10) && 20",
+			expected: NumberType{},
+		},
+		{
+			what:     "narrow type of nested && operator",
+			input:    "((('foo' && true) || false) && 10) || 20",
+			expected: NumberType{},
+		},
+		{
+			what:     "narrow type of nested || operator",
+			input:    "((('foo' || true) && false) || 10) && 20",
+			expected: NumberType{},
+		},
+		{
+			what:     "don't narrow type on nested && operator",
+			input:    "('foo' && 10) && 20",
+			expected: StringType{},
+		},
+		{
+			what:     "don't narrow type on nested || operator",
+			input:    "('foo' || 10) || 20",
+			expected: StringType{},
+		},
+		{
+			what:     "narrowed || operator at LHS and && operator at RHS",
+			input:    "('foo' || 10) && (('foo' && 10) || 20)",
+			expected: NumberType{},
+		},
+		{
+			what:     "narrowed && operator at LHS and || operator at RHS",
+			input:    "('foo' && 10) || (('foo' || 10) && 20)",
+			expected: NumberType{},
+		},
+		{
+			what:     "not operator negates && operator type narrowing",
+			input:    "!('foo' && 10) && 20",
+			expected: NumberType{},
+		},
+		{
+			what:     "not operator negates || operator type narrowing",
+			input:    "!('foo' || 10) || 20",
+			expected: NumberType{},
+		},
+		{
+			what:     "double not operators does nothing on type narrowing",
+			input:    "!!('foo' || 10) && 20",
+			expected: NumberType{},
+		},
 	}
 
 	allSPFuncs := []string{}
@@ -660,7 +734,7 @@ func TestExprSemanticsCheckOK(t *testing.T) {
 				t.Fatal("Parse error:", tc.input)
 			}
 
-			c := NewExprSemanticsChecker(false)
+			c := NewExprSemanticsChecker(false, nil)
 			if tc.funcs != nil {
 				c.funcs = tc.funcs
 			}
@@ -702,45 +776,18 @@ func TestExprSemanticsCheckOK(t *testing.T) {
 	}
 }
 
-func TestExprBuiltinFunctionSignatures(t *testing.T) {
-	for name, sigs := range BuiltinFuncSignatures {
-		if len(sigs) == 0 {
-			t.Errorf("overload candidates of %q should not be empty", name)
-		}
-		{
-			ok := true
-			for _, r := range name {
-				if !unicode.IsLower(r) {
-					ok = false
-					break
-				}
-			}
-			if !ok {
-				t.Errorf("name of function must be in lower case to check in case insensitive: %q", name)
-			}
-		}
-		for i, sig := range sigs {
-			if name != strings.ToLower(sig.Name) {
-				t.Errorf("name of %dth overload is different from its key: name=%q vs key=%q", i+1, sig.Name, name)
-			}
-			if sig.VariableLengthParams && len(sig.Params) == 0 {
-				t.Errorf("number of arguments of %dth overload of %q must not be empty because VariableLengthParams is set to true", i+1, name)
-			}
-		}
-	}
-}
-
 func TestExprSemanticsCheckError(t *testing.T) {
 	testCases := []struct {
-		what     string
-		input    string
-		expected []string
-		funcs    map[string][]*FuncSignature
-		matrix   *ObjectType
-		steps    *ObjectType
-		needs    *ObjectType
-		availCtx []string
-		availSP  []string
+		what       string
+		input      string
+		expected   []string
+		funcs      map[string][]*FuncSignature
+		matrix     *ObjectType
+		steps      *ObjectType
+		needs      *ObjectType
+		availCtx   []string
+		availSP    []string
+		configVars []string
 	}{
 		{
 			what:  "undefined variable",
@@ -969,6 +1016,13 @@ func TestExprSemanticsCheckError(t *testing.T) {
 				"takes at least 2 parameters but 1 arguments are given",
 			},
 		},
+		{
+			what:  "function name of format() call check is case insensitive",
+			input: "Format('{0}', 1, 2)",
+			expected: []string{
+				`format string "{0}" does not contain placeholder {1}`,
+			},
+		},
 		{
 			what:  "undefined matrix value",
 			input: "matrix.bar",
@@ -1147,6 +1201,36 @@ func TestExprSemanticsCheckError(t *testing.T) {
 			},
 			availSP: []string{"fail"},
 		},
+		{
+			what:  "no configuration variable is allowed",
+			input: "vars.UNKNOWN_VARIABLE",
+			expected: []string{
+				"no configuration variable is allowed since the variables list is empty",
+			},
+			configVars: []string{},
+		},
+		{
+			what:  "unknown configuration variable",
+			input: "vars.UNKNOWN_VARIABLE",
+			expected: []string{
+				"undefined configuration variable \"unknown_variable\".",
+			},
+			configVars: []string{"FOO_BAR"},
+		},
+		{
+			what:  "config variable naming convention",
+			input: "vars.FOO-BAR",
+			expected: []string{
+				"configuration variable name \"foo-bar\" can only contain alphabets, decimal numbers, and '_'.",
+			},
+		},
+		{
+			what:  "config variable name cannot start with GITHUB_",
+			input: "vars.GITHUB_FOOOOOOOO",
+			expected: []string{
+				"must not start with the GITHUB_ prefix",
+			},
+		},
 	}
 
 	allSP := []string{}
@@ -1162,7 +1246,7 @@ func TestExprSemanticsCheckError(t *testing.T) {
 				t.Fatal("Parse error:", tc.input)
 			}
 
-			c := NewExprSemanticsChecker(false)
+			c := NewExprSemanticsChecker(false, tc.configVars)
 			if tc.funcs != nil {
 				c.funcs = tc.funcs // Set functions for testing
 			}
@@ -1183,6 +1267,7 @@ func TestExprSemanticsCheckError(t *testing.T) {
 			} else {
 				c.SetSpecialFunctionAvailability(allSP)
 			}
+
 			_, errs := c.Check(e)
 			if len(errs) != len(tc.expected) {
 				t.Fatalf("semantics check should report %d errors but got %d errors: %v", len(tc.expected), len(errs), errs)
@@ -1200,8 +1285,244 @@ func TestExprSemanticsCheckError(t *testing.T) {
 	}
 }
 
+func TestExprCompareOperandsCheck(t *testing.T) {
+	// Matrix of operator -> lhs type -> rhs type -> result
+	// Result `true` means the comparison is allowed. `false` means the comparison causes an error.
+	matrix := map[string]map[string]map[string]bool{
+		"==": {
+			"number": {
+				"any":    true,
+				"number": true,
+				"string": true,
+				"bool":   true,
+				"null":   true,
+				"object": false,
+				"array":  false,
+			},
+			"string": {
+				"any":    true,
+				"number": true,
+				"string": true,
+				"bool":   true,
+				"null":   true,
+				"object": false,
+				"array":  false,
+			},
+			"bool": {
+				"any":    true,
+				"number": true,
+				"string": true,
+				"bool":   true,
+				"null":   true,
+				"object": false,
+				"array":  false,
+			},
+			"null": {
+				"any":    true,
+				"number": true,
+				"string": true,
+				"bool":   true,
+				"null":   true,
+				"object": true,
+				"array":  true,
+			},
+			"object": {
+				"any":    true,
+				"number": false,
+				"string": false,
+				"bool":   false,
+				"null":   true,
+				"object": true,
+				"array":  false,
+			},
+			"array": {
+				"any":      true,
+				"number":   false,
+				"string":   false,
+				"bool":     false,
+				"null":     true,
+				"object":   false,
+				"array":    true,
+				"array_2d": false,
+			},
+			"array_2d": {
+				"any":      true,
+				"number":   false,
+				"string":   false,
+				"bool":     false,
+				"null":     true,
+				"object":   false,
+				"array":    false,
+				"array_2d": true,
+			},
+			"any": {
+				"any":    true,
+				"number": true,
+				"string": true,
+				"bool":   true,
+				"null":   true,
+				"object": true,
+				"array":  true,
+			},
+		},
+		"<": {
+			"number": {
+				"any":    true,
+				"number": true,
+				"string": true,
+				"bool":   false,
+				"null":   false,
+				"object": false,
+				"array":  false,
+			},
+			"string": {
+				"any":    true,
+				"number": true,
+				"string": true,
+				"bool":   false,
+				"null":   false,
+				"object": false,
+				"array":  false,
+			},
+			"bool": {
+				"any":    false,
+				"number": false,
+				"string": false,
+				"bool":   false,
+				"null":   false,
+				"object": false,
+				"array":  false,
+			},
+			"null": {
+				"any":    false,
+				"number": false,
+				"string": false,
+				"bool":   false,
+				"null":   false,
+				"object": false,
+				"array":  false,
+			},
+			"any": {
+				"any":    true,
+				"number": true,
+				"string": true,
+				"bool":   false,
+				"null":   false,
+				"object": false,
+				"array":  false,
+			},
+			"object": {
+				"any":    false,
+				"number": false,
+				"string": false,
+				"bool":   false,
+				"null":   false,
+				"object": false,
+				"array":  false,
+			},
+			"array": {
+				"any":      false,
+				"number":   false,
+				"string":   false,
+				"bool":     false,
+				"null":     false,
+				"object":   false,
+				"array":    false,
+				"array_2d": false,
+			},
+		},
+	}
+
+	for op, expr := range matrix {
+		for lhs, rest := range expr {
+			for rhs, ok := range rest {
+				input := lhs + " " + op + " " + rhs
+				t.Run(input, func(t *testing.T) {
+					p := NewExprParser()
+					e, err := p.Parse(NewExprLexer(input + "}}"))
+					if err != nil {
+						t.Fatal("Parse error:", input)
+					}
+
+					c := NewExprSemanticsChecker(false, nil)
+					c.vars = map[string]ExprType{
+						"number": NumberType{},
+						"string": StringType{},
+						"bool":   BoolType{},
+						"object": NewEmptyObjectType(),
+						"array":  &ArrayType{Elem: NumberType{}},
+						"array_2d": &ArrayType{
+							Elem: &ArrayType{Elem: NumberType{}},
+						},
+						"any": AnyType{},
+					}
+
+					ty, errs := c.Check(e)
+					if ok {
+						if len(errs) > 0 {
+							t.Fatal("semantics check failed:", errs)
+						}
+						if _, ok := ty.(BoolType); !ok {
+							t.Fatalf("wanted bool type but have %q", ty)
+						}
+					} else {
+						if len(errs) != 1 {
+							t.Fatal("more than one error occurred", errs)
+						}
+						msg := errs[0].Error()
+
+						toType := func(n string) string {
+							switch n {
+							case "array":
+								return "array<number>"
+							case "array_2d":
+								return "array<array<number>>"
+							default:
+								return n
+							}
+						}
+						want := fmt.Sprintf("%q value cannot be compared to %q value with %q operator", toType(lhs), toType(rhs), op)
+
+						if !strings.Contains(msg, want) {
+							t.Fatalf("error message %q doesn't contain expected message %q", msg, want)
+						}
+					}
+				})
+			}
+		}
+	}
+}
+
+func TestExprBuiltinFunctionSignatures(t *testing.T) {
+	for name, sigs := range BuiltinFuncSignatures {
+		if len(sigs) == 0 {
+			t.Errorf("overload candidates of %q should not be empty", name)
+		}
+		{
+			ok := true
+			for _, r := range name {
+				if !unicode.IsLower(r) {
+					ok = false
+					break
+				}
+			}
+			if !ok {
+				t.Errorf("name of function must be in lower case to check in case insensitive: %q", name)
+			}
+		}
+		for i, sig := range sigs {
+			if name != strings.ToLower(sig.Name) {
+				t.Errorf("name of %dth overload is different from its key: name=%q vs key=%q", i+1, sig.Name, name)
+			}
+			if sig.VariableLengthParams && len(sig.Params) == 0 {
+				t.Errorf("number of arguments of %dth overload of %q must not be empty because VariableLengthParams is set to true", i+1, name)
+			}
+		}
+	}
+}
+
 func TestExprSemanticsCheckerUpdateMatrix(t *testing.T) {
-	c := NewExprSemanticsChecker(false)
+	c := NewExprSemanticsChecker(false, nil)
 	ty := NewEmptyObjectType()
 	prev := c.vars["matrix"]
 	c.UpdateMatrix(ty)
@@ -1216,7 +1537,7 @@ func TestExprSemanticsCheckerUpdateMatrix(t *testing.T) {
 }
 
 func TestExprSemanticsCheckerUpdateSteps(t *testing.T) {
-	c := NewExprSemanticsChecker(false)
+	c := NewExprSemanticsChecker(false, nil)
 	ty := NewEmptyObjectType()
 	prev := c.vars["steps"]
 	c.UpdateSteps(ty)
@@ -1232,7 +1553,7 @@ func TestExprSemanticsCheckerUpdateSteps(t *testing.T) {
 
 func TestExprSematincsCheckerUpdateDispatchInputsVarType(t *testing.T) {
 	ty := NewStrictObjectType(map[string]ExprType{"foo": NullType{}})
-	c := NewExprSemanticsChecker(false)
+	c := NewExprSemanticsChecker(false, nil)
 	c.UpdateDispatchInputs(ty)
 	o := c.vars["github"].(*ObjectType).Props["event"].(*ObjectType).Props["inputs"].(*ObjectType)
 	if _, ok := o.Props["foo"]; !ok {
@@ -1254,7 +1575,64 @@ func TestExprSematincsCheckerUpdateDispatchInputsVarType(t *testing.T) {
 	}
 }
 
+func TestExprSemanticsCheckerUpdateInputsMultipleTimes(t *testing.T) {
+	tests := []struct {
+		first  *ObjectType
+		second *ObjectType
+		want   *ObjectType
+	}{
+		{
+			first: NewStrictObjectType(map[string]ExprType{
+				"foo": StringType{},
+			}),
+			second: NewStrictObjectType(map[string]ExprType{
+				"bar": NumberType{},
+			}),
+			want: NewStrictObjectType(map[string]ExprType{
+				"foo": StringType{},
+				"bar": NumberType{},
+			}),
+		},
+		{
+			first: NewStrictObjectType(map[string]ExprType{
+				"foo": StringType{},
+			}),
+			second: NewStrictObjectType(map[string]ExprType{
+				"foo": StringType{},
+			}),
+			want: NewStrictObjectType(map[string]ExprType{
+				"foo": StringType{},
+			}),
+		},
+		{
+			first: NewStrictObjectType(map[string]ExprType{
+				"foo": BoolType{},
+			}),
+			second: NewStrictObjectType(map[string]ExprType{
+				"foo": NumberType{},
+			}),
+			want: NewStrictObjectType(map[string]ExprType{
+				"foo": AnyType{},
+			}),
+		},
+	}
+
+	for _, tc := range tests {
+		name := fmt.Sprintf("%v then %v", tc.first, tc.second)
+		t.Run(name, func(t *testing.T) {
+			c := NewExprSemanticsChecker(false, nil)
+			c.UpdateInputs(tc.first)
+			c.UpdateDispatchInputs(tc.second)
+			have := c.vars["inputs"]
+			if !cmp.Equal(have, tc.want) {
+				t.Fatal("Merged `inputs` type is unexpected", have, "v.s.", tc.want)
+			}
+		})
+	}
+}
+
 func testObjectPropertiesAreInLowerCase(t *testing.T, ty ExprType) {
+	t.Helper()
 	switch ty := ty.(type) {
 	case *ObjectType:
 		for n, ty := range ty.Props {
