@@ -2,6 +2,9 @@ package actionlint
 
 import (
 	"bufio"
+	"bytes"
+	"encoding/json"
+	"errors"
 	"fmt"
 	"io"
 	"os"
@@ -16,6 +19,12 @@ import (
 	"golang.org/x/sys/execabs"
 )
 
+type testErrorReader struct{}
+
+func (r testErrorReader) Read(p []byte) (int, error) {
+	return 0, errors.New("dummy read error")
+}
+
 func TestLinterLintOK(t *testing.T) {
 	dir := filepath.Join("testdata", "ok")
 
@@ -36,18 +45,18 @@ func TestLinterLintOK(t *testing.T) {
 	}
 
 	proj := &Project{root: dir}
-	shellcheck, err := execabs.LookPath("shellcheck")
-	if err != nil {
-		t.Skip("skipped because \"shellcheck\" command does not exist in system")
-	}
-
-	pyflakes, err := execabs.LookPath("pyflakes")
-	if err != nil {
-		t.Skip("skipped because \"pyflakes\" command does not exist in system")
-	}
+	shellcheck, _ := execabs.LookPath("shellcheck")
+	pyflakes, _ := execabs.LookPath("pyflakes")
 
 	for _, f := range fs {
 		t.Run(filepath.Base(f), func(t *testing.T) {
+			if strings.Contains(f, "shellcheck") && shellcheck == "" {
+				t.Skip("skipping", f, "because \"shellcheck\" command does not exist in system")
+			}
+			if strings.Contains(f, "pyflakes") && pyflakes == "" {
+				t.Skip("skipping", f, "because \"pyflakes\" command does not exist in system")
+			}
+
 			opts := LinterOptions{
 				Shellcheck: shellcheck,
 				Pyflakes:   pyflakes,
@@ -198,13 +207,72 @@ func TestLinterLintError(t *testing.T) {
 	}
 }
 
+func TestLinterLintAllErrorWorkflowsAtOnce(t *testing.T) {
+	shellcheck, err := execabs.LookPath("shellcheck")
+	if err != nil {
+		t.Skipf("shellcheck is not found: %s", err)
+	}
+
+	pyflakes, err := execabs.LookPath("pyflakes")
+	if err != nil {
+		t.Skipf("pyflakes is not found: %s", err)
+	}
+
+	dir, files, err := testFindAllWorkflowsInDir("examples")
+	if err != nil {
+		panic(err)
+	}
+
+	// Root directory must be testdata/examples/ since some workflow assumes it
+	proj := &Project{root: dir}
+
+	_, fs, err := testFindAllWorkflowsInDir("err")
+	if err != nil {
+		panic(err)
+	}
+	files = append(files, fs...)
+
+	o := LinterOptions{
+		Shellcheck: shellcheck,
+		Pyflakes:   pyflakes,
+	}
+
+	l, err := NewLinter(io.Discard, &o)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	l.defaultConfig = &Config{}
+
+	errs, err := l.LintFiles(files, proj)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	// Check each example workflow file caused at least one error
+CheckFiles:
+	for _, f := range files {
+		for _, e := range errs {
+			if e.Filepath == f {
+				continue CheckFiles
+			}
+		}
+		if !strings.Contains(f, "pyflakes") && !strings.Contains(f, "shellcheck") {
+			t.Errorf("Workflow %q caused no error: %v", f, errs)
+		}
+	}
+}
+
 func TestLintFindProjectFromPath(t *testing.T) {
-	d := filepath.Join("testdata", "projects", "broken_reusable_workflow")
-	f := filepath.Join(d, "workflows", "test.yaml")
+	d := filepath.Join("testdata", "find_project")
+	f := filepath.Join(d, ".github", "workflows", "test.yaml")
 	b, err := os.ReadFile(f)
 	if err != nil {
 		panic(err)
 	}
+
+	testEnsureDotGitDir(d)
+
 	lint := func(path string) []*Error {
 		l, err := NewLinter(io.Discard, &LinterOptions{})
 		if err != nil {
@@ -285,10 +353,6 @@ func TestLinterFormatErrorMessageOK(t *testing.T) {
 			file:   "test.jsonl",
 			format: "{{range $err := .}}{{json $err}}{{end}}",
 		},
-		{
-			file:   "test.jsonl",
-			format: "{{range $err := .}}{{json $err}}{{end}}",
-		},
 		{
 			file:   "test.md",
 			format: "{{range $ := .}}### Error at line {{$.Line}}, col {{$.Column}} of `{{$.Filepath}}`\\n\\n{{$.Message}}\\n\\n```\\n{{$.Snippet}}\\n```\\n\\n{{end}}",
@@ -339,6 +403,107 @@ func TestLinterFormatErrorMessageOK(t *testing.T) {
 	}
 }
 
+func TestLinterFormatErrorMessageInSARIF(t *testing.T) {
+	dir := filepath.Join("testdata", "format")
+	proj := &Project{root: dir}
+	file := filepath.Join(dir, "test.yaml")
+
+	bytes, err := os.ReadFile(filepath.Join(dir, "sarif_template.txt"))
+	if err != nil {
+		panic(err)
+	}
+	format := string(bytes)
+
+	opts := LinterOptions{Format: format}
+	var b strings.Builder
+	l, err := NewLinter(&b, &opts)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	l.defaultConfig = &Config{}
+	errs, err := l.LintFile(file, proj)
+	if err != nil {
+		t.Fatal(err)
+	}
+	if len(errs) == 0 {
+		t.Fatal("no error")
+	}
+
+	out := b.String()
+	// Fix path separators on Windows
+	if runtime.GOOS == "windows" {
+		slash := filepath.ToSlash(file)
+		escaped := strings.ReplaceAll(file, `\`, `\\`)
+		out = strings.ReplaceAll(out, escaped, slash)
+	}
+
+	var have interface{}
+	if err := json.Unmarshal([]byte(out), &have); err != nil {
+		t.Fatalf("output is not JSON: %v: %q", err, out)
+	}
+
+	bytes, err = os.ReadFile(filepath.Join(dir, "test.sarif"))
+	if err != nil {
+		panic(err)
+	}
+	var want interface{}
+	if err := json.Unmarshal(bytes, &want); err != nil {
+		panic(err)
+	}
+
+	if !cmp.Equal(want, have) {
+		t.Fatal(cmp.Diff(want, have))
+	}
+}
+
+func TestLinterLintStdinOK(t *testing.T) {
+	for _, f := range []string{"", "foo.yaml"} {
+		l, err := NewLinter(io.Discard, &LinterOptions{StdinFileName: f})
+		if err != nil {
+			t.Fatalf("creating Linter object with stdin file name %q caused error: %v", f, err)
+		}
+
+		in := []byte(`on: push
+jobs:
+  job:
+    runs-on: foo
+	steps:
+	  - run: echo`)
+		errs, err := l.LintStdin(bytes.NewReader(in))
+		if err != nil {
+			t.Fatalf("linting input with stdin file name %q caused error: %v", f, err)
+		}
+		if len(errs) != 1 {
+			t.Fatalf("unexpected number of errors with stdin file name %q: %v", f, errs)
+		}
+
+		want := f
+		if want == "" {
+			want = "<stdin>"
+		}
+		if errs[0].Filepath != want {
+			t.Fatalf("file path in the error with stdin file name %q should be %q but got %q", f, want, errs[0].Filepath)
+		}
+	}
+}
+
+func TestLinterLintStdinReadError(t *testing.T) {
+	l, err := NewLinter(io.Discard, &LinterOptions{})
+	if err != nil {
+		t.Fatal(err)
+	}
+	_, err = l.LintStdin(testErrorReader{})
+	if err == nil {
+		t.Fatal("error did not occur")
+	}
+	want := "could not read stdin: dummy read error"
+	have := err.Error()
+	if want != have {
+		t.Fatalf("wanted error message %q but have %q", want, have)
+	}
+}
+
 func TestLinterPathsNotFound(t *testing.T) {
 	l, err := NewLinter(io.Discard, &LinterOptions{})
 	if err != nil {
@@ -367,6 +532,130 @@ func TestLinterPathsNotFound(t *testing.T) {
 	}
 }
 
+type customRuleForTest struct {
+	RuleBase
+	count int
+}
+
+func (r *customRuleForTest) VisitStep(n *Step) error {
+	r.count++
+	if r.count > 1 {
+		r.Errorf(n.Pos, "only single step is allowed but got %d steps", r.count)
+	}
+	return nil
+}
+
+func TestLinterAddCustomRuleOnRulesCreatedHook(t *testing.T) {
+	o := &LinterOptions{
+		OnRulesCreated: func(rules []Rule) []Rule {
+			r := &customRuleForTest{
+				RuleBase: NewRuleBase("this-is-test", ""),
+			}
+			return append(rules, r)
+		},
+	}
+
+	l, err := NewLinter(io.Discard, o)
+	if err != nil {
+		t.Fatal(err)
+	}
+	l.defaultConfig = &Config{}
+
+	{
+		w := `on: push
+jobs:
+  test:
+    runs-on: ubuntu-latest
+    steps:
+      - run: echo
+`
+		errs, err := l.Lint("test.yaml", []byte(w), nil)
+		if err != nil {
+			t.Fatal(err)
+		}
+		if len(errs) != 0 {
+			t.Fatal("wanted no error but have", errs)
+		}
+	}
+
+	{
+		w := `on: push
+jobs:
+  test:
+    runs-on: ubuntu-latest
+    steps:
+      - run: echo
+      - run: echo
+`
+		errs, err := l.Lint("test.yaml", []byte(w), nil)
+		if err != nil {
+			t.Fatal(err)
+		}
+		if len(errs) != 1 {
+			t.Fatal("wanted 1 error but have", errs)
+		}
+
+		var b strings.Builder
+		errs[0].PrettyPrint(&b, nil)
+		have := b.String()
+		want := "test.yaml:7:9: only single step is allowed but got 2 steps [this-is-test]\n"
+		if have != want {
+			t.Fatalf("wanted error message %q but have %q", want, have)
+		}
+	}
+}
+
+func TestLinterRemoveRuleOnRulesCreatedHook(t *testing.T) {
+	o := &LinterOptions{
+		OnRulesCreated: func(rules []Rule) []Rule {
+			for i, r := range rules {
+				if r.Name() == "runner-label" {
+					rules = append(rules[:i], rules[i+1:]...)
+					break
+				}
+			}
+			return rules
+		},
+	}
+
+	l, err := NewLinter(io.Discard, o)
+	if err != nil {
+		t.Fatal(err)
+	}
+	l.defaultConfig = &Config{}
+
+	f := filepath.Join("testdata", "err", "invalid_runner_labels.yaml")
+	errs, err := l.LintFile(f, nil)
+	if err != nil {
+		t.Fatal(err)
+	}
+	if len(errs) != 0 {
+		t.Fatal("no error was expected because runner-label rule was removed but got:", errs)
+	}
+}
+
+func TestLinterGenerateDefaultConfigAlreadyExists(t *testing.T) {
+	l, err := NewLinter(io.Discard, &LinterOptions{})
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	for _, n := range []string{"ok", "yml"} {
+		d := filepath.Join("testdata", "config", "projects", n)
+		testEnsureDotGitDir(d)
+
+		err := l.GenerateDefaultConfig(d)
+		if err == nil {
+			t.Fatal("error did not occur for project", d)
+		}
+		want := "config file already exists at"
+		msg := err.Error()
+		if !strings.Contains(msg, want) {
+			t.Fatalf("error message %q does not contain expected text %q", msg, want)
+		}
+	}
+}
+
 func BenchmarkLintWorkflowFiles(b *testing.B) {
 	large := filepath.Join("testdata", "bench", "many_scripts.yaml")
 	small := filepath.Join("testdata", "bench", "small.yaml")
