@@ -11,8 +11,7 @@ const (
 	compatInvalid                   = 0
 	compatUbuntu2004 runnerOSCompat = 1 << iota
 	compatUbuntu2204
-	compatMacOS1015
-	compatMacOS110
+	compatUbuntu2404
 	compatMacOS120
 	compatMacOS120L
 	compatMacOS120XL
@@ -22,7 +21,9 @@ const (
 	compatMacOS140
 	compatMacOS140L
 	compatMacOS140XL
-	compatWindows2016
+	compatMacOS150
+	compatMacOS150L
+	compatMacOS150XL
 	compatWindows2019
 	compatWindows2022
 )
@@ -33,35 +34,32 @@ var allGitHubHostedRunnerLabels = []string{
 	"windows-latest-8-cores",
 	"windows-2022",
 	"windows-2019",
-	"windows-2016",
 	"ubuntu-latest",
 	"ubuntu-latest-4-cores",
 	"ubuntu-latest-8-cores",
 	"ubuntu-latest-16-cores",
+	"ubuntu-24.04",
 	"ubuntu-22.04",
 	"ubuntu-20.04",
 	"macos-latest",
 	"macos-latest-xl",
 	"macos-latest-xlarge",
 	"macos-latest-large",
+	"macos-15-xlarge",
+	"macos-15-large",
+	"macos-15",
 	"macos-14-xl",
 	"macos-14-xlarge",
 	"macos-14-large",
 	"macos-14",
-	"macos-14.0",
 	"macos-13-xl",
 	"macos-13-xlarge",
 	"macos-13-large",
 	"macos-13",
-	"macos-13.0",
 	"macos-12-xl",
 	"macos-12-xlarge",
 	"macos-12-large",
 	"macos-12",
-	"macos-12.0",
-	"macos-11",
-	"macos-11.0",
-	"macos-10.15",
 }
 
 // https://docs.github.com/en/actions/hosting-your-own-runners/using-self-hosted-runners-in-a-workflow#using-default-labels-to-route-jobs
@@ -84,38 +82,35 @@ var defaultRunnerOSCompats = map[string]runnerOSCompat{
 	"ubuntu-latest-4-cores":  compatUbuntu2204,
 	"ubuntu-latest-8-cores":  compatUbuntu2204,
 	"ubuntu-latest-16-cores": compatUbuntu2204,
+	"ubuntu-24.04":           compatUbuntu2404,
 	"ubuntu-22.04":           compatUbuntu2204,
 	"ubuntu-20.04":           compatUbuntu2004,
 	"macos-latest-xl":        compatMacOS140XL,
 	"macos-latest-xlarge":    compatMacOS140XL,
 	"macos-latest-large":     compatMacOS140L,
 	"macos-latest":           compatMacOS140,
+	"macos-15-xlarge":        compatMacOS150XL,
+	"macos-15-large":         compatMacOS150L,
+	"macos-15":               compatMacOS150,
 	"macos-14-xl":            compatMacOS140XL,
 	"macos-14-xlarge":        compatMacOS140XL,
 	"macos-14-large":         compatMacOS140L,
 	"macos-14":               compatMacOS140,
-	"macos-14.0":             compatMacOS140,
 	"macos-13-xl":            compatMacOS130XL,
 	"macos-13-xlarge":        compatMacOS130XL,
 	"macos-13-large":         compatMacOS130L,
 	"macos-13":               compatMacOS130,
-	"macos-13.0":             compatMacOS130,
 	"macos-12-xl":            compatMacOS120XL,
 	"macos-12-xlarge":        compatMacOS120XL,
 	"macos-12-large":         compatMacOS120L,
 	"macos-12":               compatMacOS120,
-	"macos-12.0":             compatMacOS120,
-	"macos-11":               compatMacOS110,
-	"macos-11.0":             compatMacOS110,
-	"macos-10.15":            compatMacOS1015,
 	"windows-latest":         compatWindows2022,
 	"windows-latest-8-cores": compatWindows2022,
 	"windows-2022":           compatWindows2022,
 	"windows-2019":           compatWindows2019,
-	"windows-2016":           compatWindows2016,
-	"linux":                  compatUbuntu2204 | compatUbuntu2004, // Note: "linux" does not always indicate Ubuntu. It might be Fedora or Arch or ...
-	"macos":                  compatMacOS140 | compatMacOS140L | compatMacOS140XL | compatMacOS130 | compatMacOS130L | compatMacOS130XL | compatMacOS120 | compatMacOS120L | compatMacOS120XL | compatMacOS110 | compatMacOS1015,
-	"windows":                compatWindows2022 | compatWindows2019 | compatWindows2016,
+	"linux":                  compatUbuntu2404 | compatUbuntu2204 | compatUbuntu2004, // Note: "linux" does not always indicate Ubuntu. It might be Fedora or Arch or ...
+	"macos":                  compatMacOS150 | compatMacOS150L | compatMacOS150XL | compatMacOS140 | compatMacOS140L | compatMacOS140XL | compatMacOS130 | compatMacOS130L | compatMacOS130XL | compatMacOS120 | compatMacOS120L | compatMacOS120XL,
+	"windows":                compatWindows2022 | compatWindows2019,
 }
 
 // RuleRunnerLabel is a rule to check runner label like "ubuntu-latest". There are two types of
